import os
import logging
from tokenizers import BertWordPieceTokenizer, ByteLevelBPETokenizer
from transformers import (
    WEIGHTS_NAME,
    AutoConfig,
    AutoModelWithLMHead,
    AutoTokenizer,
    BertConfig,
    BertForMaskedLM,
    BertTokenizer,
    BigBirdConfig,
    BigBirdForMaskedLM,
    BigBirdTokenizer,
    CamembertConfig,
    CamembertForMaskedLM,
    CamembertTokenizer,
    DistilBertConfig,
    DistilBertForMaskedLM,
    DistilBertTokenizer,
    ElectraConfig,
    ElectraForMaskedLM,
    ElectraForPreTraining,
    ElectraTokenizer,
    GPT2Config,
    GPT2LMHeadModel,
    GPT2Tokenizer,
    LongformerConfig,
    LongformerForMaskedLM,
    LongformerTokenizer,
    OpenAIGPTConfig,
    OpenAIGPTLMHeadModel,
    OpenAIGPTTokenizer,
    PreTrainedModel,
    PreTrainedTokenizer,
    RobertaConfig,
    RobertaForMaskedLM,
    RobertaTokenizer,
    XLMRobertaConfig,
    XLMRobertaForMaskedLM,
    XLMRobertaTokenizer,
)

log = logging.getLogger(__name__)

MODEL_CLASSES = {
    "auto": (AutoConfig, AutoModelWithLMHead, AutoTokenizer),
    "bert": (BertConfig, BertForMaskedLM, BertTokenizer),
    "distilbert": (DistilBertConfig, DistilBertForMaskedLM, DistilBertTokenizer),
    "electra": (ElectraConfig, ElectraForMaskedLM, ElectraTokenizer),
    "gpt2": (GPT2Config, GPT2LMHeadModel, GPT2Tokenizer),
}


def train_tokenizer(
    train_files,
    args,
    tokenizer_name=None,
    output_dir=None,
):
    """
    Train a new tokenizer on `train_files`.

    Args:

    - train_files: List of files to be used when training the tokenizer.

    - tokenizer_name: Name of a pretrained tokenizer or a path to a directory containing a tokenizer.

    - output_dir (optional): The directory where model files will be saved. If not given, args.output_dir
    will be used.

    Returns: None
    """

    if not args.vocab_size:
        raise AttributeError(
            "Cannot train a new tokenizer as vocab_size is not specified in args dict. "
            "Either provide a tokenizer or specify vocab_size."
        )

    if not isinstance(train_files, list):
        train_files = [train_files]

    if not output_dir:
        output_dir = args.output_dir

    if args.model_type in ["bert", "electra"]:
        tokenizer = BertWordPieceTokenizer(
            clean_text=args.clean_text,
            handle_chinese_chars=args.handle_chinese_chars,
            strip_accents=args.strip_accents,  # Must be False if cased model
            lowercase=args.do_lower_case,
        )
        # args.special_tokens = ["[PAD]", "[UNK]", "[CLS]", "[SEP]", "[MASK]"]
        # args.wordpieces_prefix = "##"
        unused_tokens = [f"[unused{n}]" for n in range(args.unused_token_num)]
        special_tokens = list(args.special_tokens) + unused_tokens

        tokenizer.train(
            files=train_files,
            vocab_size=args.vocab_size,
            limit_alphabet=args.limit_alphabet,
            min_frequency=args.min_frequency,
            special_tokens=special_tokens,
            wordpieces_prefix=args.wordpieces_prefix,
        )
    elif args.model_type in ["bigbird", "xlmroberta"]:
        # The google BigBird way
        # Tokenizers sentencepiece does not build a BigBird compatible vocabulary model
        import sentencepiece as spm
        import shutil

        os.makedirs(output_dir, exist_ok=True)
        files = ",".join(train_files)

        if args.model_type in ["xlmroberta"]:
            # </s>,<s>,<unk>,<pad> are built in -- leave as default
            # XLMRoberta uses sentencepiece.bpe as a vocab model prefix
            prefix = "sentencepiece.bpe"
            spm.SentencePieceTrainer.Train(
                f"--input={files} --user_defined_symbols='<mask>,<s>NOTUSED,</s>NOTUSED' --model_prefix={prefix} --vocab_size={args.vocab_size - 2}"
            )
        else:
            # </s>,<s>,<unk>,<pad> are built in -- leave as default
            # BigBird uses spiece as a vocab model prefix
            prefix = "spiece"
            spm.SentencePieceTrainer.Train(
                f"--input={files} --user_defined_symbols='[SEP],[CLS],[MASK]' --model_prefix=spiece --vocab_size={args.vocab_size - 3}"
            )

        # SentencePiece There is no option for output path https://github.com/google/sentencepiece/blob/master/doc/options.md
        if os.path.exists(output_dir + "/" + f"{prefix}.model"):
            os.remove(output_dir + "/" + f"{prefix}.model")
        shutil.move(src=f"{prefix}.model", dst=output_dir)

        if os.path.exists(output_dir + "/" + f"{prefix}.vocab"):
            os.remove(output_dir + "/" + f"{prefix}.vocab")
        shutil.move(src=f"{prefix}.vocab", dst=output_dir)
    else:
        tokenizer = ByteLevelBPETokenizer(lowercase=args.do_lower_case)

        tokenizer.train(
            files=train_files,
            vocab_size=args.vocab_size,
            min_frequency=args.min_frequency,
            special_tokens=list(args.special_tokens),
        )

    if args.model_type not in ["bigbird", "xlmroberta"]:
        os.makedirs(output_dir, exist_ok=True)

        tokenizer.save_model(output_dir)
        log.info(
            " Training of {} tokenizer complete. Saved to {}.".format(
                tokenizer_name, output_dir
            )
        )

    # _, _, tokenizer_class = MODEL_CLASSES[args.model_type]
    # tokenizer = tokenizer_class.from_pretrained(output_dir)


def train_spm(
    train_files,
    output_dir=None,
    train_args=None,
    **kwargs,
):
    import sentencepiece as spm
    import shutil

    os.makedirs(output_dir, exist_ok=True)
    files = ",".join(train_files)

    if args.model_type in ["xlmroberta"]:
        # </s>,<s>,<unk>,<pad> are built in -- leave as default
        # XLMRoberta uses sentencepiece.bpe as a vocab model prefix
        prefix = "sentencepiece.bpe"
        spm.SentencePieceTrainer.Train(
            f"--input={files} --user_defined_symbols='<mask>,<s>NOTUSED,</s>NOTUSED' --model_prefix={prefix} --vocab_size={args.vocab_size - 2}"
        )
    else:
        # </s>,<s>,<unk>,<pad> are built in -- leave as default
        # BigBird uses spiece as a vocab model prefix
        prefix = "spiece"
        spm.SentencePieceTrainer.Train(
            f"--input={files} --user_defined_symbols='[SEP],[CLS],[MASK]' --model_prefix=spiece --vocab_size={args.vocab_size - 3}"
        )

    # SentencePiece There is no option for output path https://github.com/google/sentencepiece/blob/master/doc/options.md
    if os.path.exists(output_dir + "/" + f"{prefix}.model"):
        os.remove(output_dir + "/" + f"{prefix}.model")
    shutil.move(src=f"{prefix}.model", dst=output_dir)

    if os.path.exists(output_dir + "/" + f"{prefix}.vocab"):
        os.remove(output_dir + "/" + f"{prefix}.vocab")
    shutil.move(src=f"{prefix}.vocab", dst=output_dir)



"""the_pile dataset"""

import io
import os
import pandas as pd
from ekorpkit import eKonf
from ekorpkit.io.fetch.web import web_download
from tqdm.auto import tqdm

try:
    import simdjson as json
except ImportError:
    print("Installing simdjson library")
    os.system("pip install -q pysimdjson")
    import json as json


parser = json.Parser()


def json_parser(x):
    try:
        line = parser.parse(x).as_dict()
        return line
    except ValueError:
        return x


class PileReader:
    def __init__(self, filenames, subset=None, segment_separator="\n\n"):
        if not isinstance(filenames, list):
            filenames = [filenames]
        self.filenames = filenames
        self.subset = subset
        self.segment_separator = segment_separator

    def _read_fn(self):
        import zstandard
        import jsonlines

        for filename in self.filenames:
            print(f"iterating over {filename}")
            with open(filename, "rb") as f:
                cctx = zstandard.ZstdDecompressor()
                reader_stream = io.BufferedReader(cctx.stream_reader(f))
                reader = jsonlines.Reader(reader_stream, loads=json_parser)
                for i, item in enumerate(reader):
                    result = dict()
                    if isinstance(item, str):
                        result["text"] = item
                        result["subset"] = "the_pile"
                    else:
                        text = item["text"]
                        if isinstance(text, list):
                            text = self.segment_separator.join(text)
                        result["text"] = text
                        result["subset"] = item.get("meta", {}).get(
                            "pile_set_name", "the_pile"
                        )
                    if self.subset is None:
                        yield result
                    else:
                        if self.subset == result["subset"]:
                            yield result

    def __iter__(self):
        return self._read_fn()


class ThePile:
    def __init__(self, **args):
        self.cfg = eKonf.to_config(args)
        self.name = self.cfg.name
        self.subsets = self.cfg.get("subsets", [])
        if self.name == "the_pile":
            self.subset = None
        else:
            self.subset = self.name
            if self.subset not in self.subsets:
                raise ValueError(f"{self.subset} not in {self.subsets}")
        self._parse_split_urls()

    def _parse_split_urls(self):
        self.splits = {}
        for split, info in self.cfg.data_sources.items():
            if info.get("splits", None):
                urls = [
                    info["url"].format(str(i).zfill(info.zfill))
                    for i in range(info.splits)
                ]
            else:
                urls = [info["url"]]
            paths = {}
            for url in urls:
                path = os.path.join(self.cfg.data_dir, url.split("/")[-1])
                paths[path] = url
            self.splits[split] = paths

    def download(self):
        for split, paths in self.splits.items():
            for path, url in paths.items():
                print(f"Downloading {split} from {url} to {path}")
                web_download(url, path, self.name)

    def load(self, split="train"):
        paths = list(self.splits[split].keys())
        return self._generate_examples(paths)

    def _generate_examples(self, paths):
        pipeline = PileReader(paths, self.subset)
        for result in pipeline:
            if result:
                yield result


def load_pile_data(split_name, **args):
    pile = ThePile(**args)

    ds_iter = pile.load(split_name)
    documents = []
    for sample in tqdm(ds_iter):
        documents.append(sample)
    df = pd.DataFrame(documents)
    return df



#!python
from __future__ import with_statement
from __future__ import division
from __future__ import absolute_import
from __future__ import print_function
from .batcher import Batcher


def decorator_apply_batch(func, batcher=None):
    def wrapper_func(*args, **kwargs):
        return ApplyBatch(
            func, args=args[1:], kwargs=kwargs, batcher=batcher
        ).transform(args[0])

    return wrapper_func


def batch_transform(args):
    f = args[1]
    f_args = args[2]
    f_kwargs = args[3]
    return f(args[0], *f_args, **f_kwargs)


class ApplyBatch(object):
    # Applies a function to the entire minibatch. Use this for example on Pandas dataframes, to avoid per-row overhead.
    # Function needs to be applicable to the array/list of values!
    # If not, modify/wrap the function to process a list, or use Apply
    def __init__(self, function, batcher=None, args=[], kwargs={}):
        if batcher is None:
            self.batcher = Batcher()
        else:
            self.batcher = batcher
        self.function = function
        self.args = [args]
        self.kwargs = [kwargs]

    def fit(self, data, input_split=False):
        return self

    def fit_transform(
        self,
        data,
        input_split=False,
        merge_output=True,
        minibatch_size=None,
        batcher=None,
    ):
        return self.transform(data, input_split, merge_output, minibatch_size, batcher)

    def transform(
        self,
        data,
        input_split=False,
        merge_output=True,
        minibatch_size=None,
        batcher=None,
    ):
        if batcher is None:
            batcher = self.batcher
        return batcher.process_batches(
            batch_transform,
            data,
            [self.function] + self.args + self.kwargs,
            input_split=input_split,
            merge_output=merge_output,
            minibatch_size=minibatch_size,
        )


# import wordbatch.batcher as batcher
# b= batcher.Batcher(minibatch_size=2)#, method="serial")
# import numpy as np
# a= ApplyBatch(np.power, b, [2],{})
# print(a.transform([1, 2, 3, 4]))



import logging
import unicodedata
import re
from ftfy import TextFixerConfig, fix_text
from .hanja import translate as hanja2hangle
from .hangle import compose, decompose


log = logging.getLogger(__name__)


doublespace_pattern = re.compile(r"\s+")
repeatchars_pattern = re.compile(r"(\w)\\1{2,}")
number_pattern = re.compile(r"[0-9]")
punctuation_pattern = re.compile(r"[,\.\?\!]")
symbol_pattern = re.compile(r"[()\[\]\{\}`]")
hangle_pattern = re.compile(r"[ㄱ-ㅎㅏ-ㅣ가-힣]")
alphabet_pattern = re.compile(r"[a-zA-Z]")

hangle_filter = re.compile(r"[^ㄱ-ㅎㅏ-ㅣ가-힣]")
hangle_number_filter = re.compile(r"[^ㄱ-ㅎㅏ-ㅣ가-힣0-9]")
text_filter = re.compile(r"[^ㄱ-ㅎㅏ-ㅣ가-힣a-zA-Z0-9,\.\?\!&·\"'\(\)\[\]\{\}+\-\\\/\*×%]")
# text_filter = re.compile(r'[^ㄱ-ㅎㅏ-ㅣ가-힣a-zA-Z0-9,\.\?\!&·\"\'-()\[\]\{\}]')
text_filter_for_lrgraph = re.compile(r"[^ㄱ-ㅎㅏ-ㅣ가-힣a-zA-Z0-9,\.\?\!&\"'-()\[\]\{\}]")


#: Control characters.
CONTROLS = {
    "\u0001",
    "\u0002",
    "\u0003",
    "\u0004",
    "\u0005",
    "\u0006",
    "\u0007",
    "\u0008",
    "\u000e",
    "\u000f",
    "\u0011",
    "\u0012",
    "\u0013",
    "\u0014",
    "\u0015",
    "\u0016",
    "\u0017",
    "\u0018",
    "\u0019",
    "\u001a",
    "\u001b",
}
# There are further control characters, but they are instead replaced with a space by unicode normalization
# '\u0009', '\u000a', '\u000b', '\u000c', '\u000d', '\u001c',  '\u001d', '\u001e', '\u001f'


#: Hyphen and dash characters.
HYPHENS = {
    "-",  # \u002d Hyphen-minus
    "‐",  # \u2010 Hyphen
    "‑",  # \u2011 Non-breaking hyphen
    "⁃",  # \u2043 Hyphen bullet
    "‒",  # \u2012 figure dash
    "–",  # \u2013 en dash
    "—",  # \u2014 em dash
    "―",  # \u2015 horizontal bar
}

#: Minus characters.
MINUSES = {
    "-",  # \u002d Hyphen-minus
    "−",  # \u2212 Minus
    "－",  # \uff0d Full-width Hyphen-minus
    "⁻",  # \u207b Superscript minus
}

#: Plus characters.
PLUSES = {
    "+",  # \u002b Plus
    "＋",  # \uff0b Full-width Plus
    "⁺",  # \u207a Superscript plus
}

#: Slash characters.
SLASHES = {
    "/",  # \u002f Solidus
    "⁄",  # \u2044 Fraction slash
    "∕",  # \u2215 Division slash
}

#: Tilde characters.
TILDES = {
    "~",  # \u007e Tilde
    "˜",  # \u02dc Small tilde
    "⁓",  # \u2053 Swung dash
    "∼",  # \u223c Tilde operator
    "∽",  # \u223d Reversed tilde
    "∿",  # \u223f Sine wave
    "〜",  # \u301c Wave dash
    "～",  # \uff5e Full-width tilde
}

#: Apostrophe characters.
APOSTROPHES = {
    "'",  # \u0027
    "’",  # \u2019
    "՚",  # \u055a
    "Ꞌ",  # \ua78b
    "ꞌ",  # \ua78c
    "＇",  # \uff07
}

#: Single quote characters.
SINGLE_QUOTES = {
    "'",  # \u0027
    "‘",  # \u2018
    "’",  # \u2019
    "‚",  # \u201a
    "‛",  # \u201b
}

#: Double quote characters.
DOUBLE_QUOTES = {
    '"',  # \u0022
    "“",  # \u201c
    "”",  # \u201d
    "„",  # \u201e
    "‟",  # \u201f
}

#: Accent characters.
ACCENTS = {
    "`",  # \u0060
    "´",  # \u00b4
}

#: Prime characters.
PRIMES = {
    "′",  # \u2032
    "″",  # \u2033
    "‴",  # \u2034
    "‵",  # \u2035
    "‶",  # \u2036
    "‷",  # \u2037
    "⁗",  # \u2057
}

#: Quote characters, including apostrophes, single quotes, double quotes, accents and primes.
QUOTES = APOSTROPHES | SINGLE_QUOTES | DOUBLE_QUOTES | ACCENTS | PRIMES

#: Uppercase and lowercase greek letters.
GREEK = {
    "Α",  # \u0391
    "Β",  # \u0392
    "Γ",  # \u0393
    "Δ",  # \u0394
    "Ε",  # \u0395
    "Ζ",  # \u0396
    "Η",  # \u0397
    "Θ",  # \u0398
    "Ι",  # \u0399
    "Κ",  # \u039a
    "Λ",  # \u039b
    "Μ",  # \u039c
    "Ν",  # \u039d
    "Ξ",  # \u039e
    "Ο",  # \u039f
    "Π",  # \u03a0
    "Ρ",  # \u03a1
    "Σ",  # \u03a3
    "Τ",  # \u03a4
    "Υ",  # \u03a5
    "Φ",  # \u03a6
    "Χ",  # \u03a7
    "Ψ",  # \u03a8
    "Ω",  # \u03a9
    "α",  # \u03b1
    "β",  # \u03b2
    "γ",  # \u03b3
    "δ",  # \u03b4
    "ε",  # \u03b5
    "ζ",  # \u03b6
    "η",  # \u03b7
    "θ",  # \u03b8
    "ι",  # \u03b9
    "κ",  # \u03ba
    "λ",  # \u03bb
    "μ",  # \u03bc
    "ν",  # \u03bd
    "ξ",  # \u03be
    "ο",  # \u03bf
    "π",  # \u03c0
    "ρ",  # \u03c1
    "σ",  # \u03c3
    "τ",  # \u03c4
    "υ",  # \u03c5
    "φ",  # \u03c6
    "χ",  # \u03c7
    "ψ",  # \u03c8
    "ω",  # \u03c9
}

#: Names of greek letters spelled out as words.
GREEK_WORDS = {
    "Alpha",
    "Beta",
    "Gamma",
    "Delta",
    "Epsilon",
    "Zeta",
    "Eta",
    "Theta",
    "Iota",
    "Kappa",
    "Lambda",
    "Mu",
    "Nu",
    "Xi",
    "Omicron",
    "Pi",
    "Rho",
    "Sigma",
    "Tau",
    "Upsilon",
    "Phi",
    "Chi",
    "Psi",
    "Omega",
    "alpha",
    "beta",
    "gamma",
    "delta",
    "epsilon",
    "zeta",
    "eta",
    "theta",
    "iota",
    "kappa",
    "lamda",
    "mu",
    "nu",
    "xi",
    "omicron",
    "pi",
    "rho",
    "sigma",
    "tau",
    "upsilon",
    "phi",
    "chi",
    "psi",
    "omega",
}

#: Words that should not be capitalized in titles.
SMALL = {
    "a",
    "an",
    "and",
    "as",
    "at",
    "but",
    "by",
    "en",
    "for",
    "if",
    "in",
    "of",
    "on",
    "or",
    "the",
    "to",
    "v",
    "v",
    "via",
    "vs",
    "vs",
}

#: Words that should not be capitalized in names.
NAME_SMALL = {
    "abu",
    "bon",
    "bin",
    "da",
    "dal",
    "de",
    "del",
    "der",
    "de",
    "di",
    u"dí",
    "ibn",
    "la",
    "le",
    "san",
    "st",
    "ste",
    "van",
    "vel",
    "von",
    "y",
}

# This isn't every possible TLD, just the most common, to avoid false positives.
TLDS = {
    "aero",
    "asia",
    "biz",
    "cat",
    "com",
    "coop",
    "edu",
    "eu",
    "gov",
    "info",
    "int",
    "jobs",
    "mil",
    "mobi",
    "museum",
    "name",
    "net",
    "org",
    "pro",
    "tel",
    "travel",
    "xxx",
    "ad",
    "as",
    "ar",
    "au",
    "br",
    "bz",
    "ca",
    "cc",
    "cd",
    "co",
    "ch",
    "cn",
    "de",
    "dj",
    "es",
    "fr",
    "fm",
    "it",
    "io",
    "jp",
    "la",
    "ly",
    "me",
    "ms",
    "nl",
    "no",
    "nu",
    "ru",
    "sc",
    "se",
    "sr",
    "su",
    "tk",
    "tv",
    "uk",
    "us",
    "ws",
}

#: A variety of numbers, spelled out as words.
NUMBERS = {
    "zero",
    "one",
    "two",
    "three",
    "four",
    "five",
    "six",
    "seven",
    "eight",
    "nine",
    "ten",
    "eleven",
    "twelve",
    "thirteen",
    "fourteen",
    "fifteen",
    "sixteen",
    "seventeen",
    "eighteen",
    "nineteen",
    "twenty",
    "thirty",
    "forty",
    "fifty",
    "sixty",
    "seventy",
    "eighty",
    "ninety",
    "hundred",
    "thousand",
    "million",
    "billion",
    "trillion",
}
LEFT_PARENTHESES = {"(", "[", "{", "&lt;"}
RIGHT_PARENTHESES = {")", "]", "}", "&gt;"}
#: Regular expression that matches email addresses.
EMAIL_RE = re.compile(r"([\w\-\.\+%]+@(\w[\w\-]+\.)+[\w\-]+)", re.I | re.U)
#: Regular expression that matches DOIs.
DOI_RE = re.compile(r"^10\.\d{4,9}/[-\._;()/:A-Z0-9]+$", re.U)
#: Regular expression that matches ISSNs.
ISSN_RE = re.compile(r"^\d{4}-\d{3}[\dX]$", re.U)
#: Regular expression that matches control characters not allowed in XML.
CONTROL_RE = re.compile(
    "[^\u0020-\uD7FF\u0009\u000A\u000D\uE000-\uFFFD\u10000-\u10FFFF]+"
)


# def get_encoding(input_string, guesses=None, is_html=False):
#     """Return the encoding of a byte string. Uses bs4 UnicodeDammit.
#     :param string input_string: Encoded byte string.
#     :param list[string] guesses: (Optional) List of encoding guesses to prioritize.
#     :param bool is_html: Whether the input is HTML.
#     """
#     converted = UnicodeDammit(input_string, override_encodings=[guesses] if guesses else [], is_html=is_html)
#     return converted.original_encoding


def levenshtein(s1, s2, allow_substring=False):
    """Return the Levenshtein distance between two strings.
    The Levenshtein distance (a.k.a "edit difference") is the number of characters that need to be substituted,
    inserted or deleted to transform s1 into s2.
    Setting the `allow_substring` parameter to True allows s1 to be a
    substring of s2, so that, for example, "hello" and "hello there" would have a distance of zero.
    :param string s1: The first string
    :param string s2: The second string
    :param bool allow_substring: Whether to allow s1 to be a substring of s2
    :returns: Levenshtein distance.
    :rtype int
    """
    len1, len2 = len(s1), len(s2)
    lev = []
    for i in range(len1 + 1):
        lev.append([0] * (len2 + 1))
    for i in range(len1 + 1):
        lev[i][0] = i
    for j in range(len2 + 1):
        lev[0][j] = 0 if allow_substring else j
    for i in range(len1):
        for j in range(len2):
            lev[i + 1][j + 1] = min(
                lev[i][j + 1] + 1, lev[i + 1][j] + 1, lev[i][j] + (s1[i] != s2[j])
            )
    return min(lev[len1]) if allow_substring else lev[len1][len2]


def bracket_level(text, open={"(", "[", "{"}, close={")", "]", "}"}):
    """Return 0 if string contains balanced brackets or no brackets."""
    level = 0
    for c in text:
        if c in open:
            level += 1
        elif c in close:
            level -= 1
    return level


def is_punct(text):
    for char in text:
        if not unicodedata.category(char).startswith("P"):
            return False
    else:
        return True


def is_ascii(text):
    for char in text:
        if ord(char) >= 128:
            return False
    else:
        return True


def like_url(text):
    if len(text) < 1:
        return False
    if text.startswith("http://"):
        return True
    elif text.startswith("www.") and len(text) >= 5:
        return True
    if len(text) < 2 or text[0] == "." or text[-1] == "." or "." not in text:
        return False
    tld = text.rsplit(".", 1)[1].split(":", 1)[0]
    if tld.endswith("/"):
        return True
    if tld.isalpha() and tld in TLDS:
        return True
    return False


def like_number(text):
    text = text.replace(",", "").replace(".", "")
    if text.isdigit():
        return True
    if text.count("/") == 1:
        num, denom = text.split("/")
        if like_number(num) and like_number(denom):
            return True
    if text in NUMBERS:
        return True
    return False


def word_shape(text):
    prev_m = ""
    seq = 0
    shape = []
    for c in text:
        if c.isdigit():
            m = "d"  # Digits
        elif c in GREEK:
            m = "g"  # Greek letters
        elif c.isalpha():
            m = "X" if c.isupper() else "x"  # Uppercase or lowercase alphabetical
        elif c in QUOTES:
            m = "'"  # Quotes and apostrophes
        elif c in {":", ";"}:
            m = ":"  # Colons and semicolons
        elif c in {"!", "?", "."}:
            m = "."  # Sentence ends
        elif c in {"(", "[", "{", ")", "]", "}"}:
            m = "b"  # Brackets
        elif c in {"°", "%"}:
            m = "u"  # units
        elif c in {"■", "◼", "●", "▲", "○", "◆", "▼", "⧫", "△", "◇", "▽", "⬚", "□"}:
            m = "l"  # list markers
        elif c in {",", "$", "&", "-"}:
            m = c  # Stay the same
        else:
            m = "*"  # Everything else, symbols etc: {'=', '+', '*', '_', '|', '@', '×', '÷', '±', '<', '≤', '>', '≥', '≦', '≡', '≅', '≈', '≃', '≲', '→', '←', '⇄', '≪', '≫', '↔', '≠', '∝', '∈', '⇌', '⇋', '⋯', '~', '·', '•', '√', '⊃', '∑', '∏', '®', '∞', '∂', '∫', '∇', '∧', '⟨', '⟩'}
        if m == prev_m:
            seq += 1
        else:
            seq = 0
            prev_m = m
        if seq < 3:
            shape.append(m)
    return "".join(shape)


def remove_doublespace(sent):
    return doublespace_pattern.sub(" ", sent)


def repeat_normalize(sent, num_repeats=2):
    if num_repeats > 0:
        sent = repeatchars_pattern.sub("\\1" * num_repeats, sent)
    sent = doublespace_pattern.sub(" ", sent)
    return sent.strip()


def emoticon_normalize(sent, num_repeats=2):
    if not sent:
        return sent

    # Pattern matching ㅋ쿠ㅜ
    def pattern(idx):
        # Jaum: 0, Moum: 1, Complete: 2, else -1
        if 12593 <= idx <= 12622:
            return 0
        elif 12623 <= idx <= 12643:
            return 1
        elif 44032 <= idx <= 55203:
            return 2
        else:
            return -1

    idxs = [pattern(ord(c)) for c in sent]
    sent_ = []
    last_idx = len(idxs) - 1
    for i, (idx, c) in enumerate(zip(idxs, sent)):
        if (i > 0 and i < last_idx) and (
            idxs[i - 1] == 0 and idx == 2 and idxs[i + 1] == 1
        ):
            cho, jung, jong = decompose(c)
            if (cho == sent[i - 1]) and (jung == sent[i + 1]) and (jong == " "):
                sent_.append(cho)
                sent_.append(jung)
            else:
                sent_.append(c)
        elif (i < last_idx) and (idx == 2) and (idxs[i + 1] == 0):
            cho, jung, jong = decompose(c)
            if jong == sent[i + 1]:
                sent_.append(compose(cho, jung, " "))
                sent_.append(jong)
        elif (i > 0) and (idx == 2 and idxs[i - 1] == 0):
            cho, jung, jong = decompose(c)
            if cho == sent[i - 1]:
                sent_.append(cho)
                sent_.append(jung)
        else:
            sent_.append(c)
    return repeat_normalize("".join(sent_), num_repeats)


def hanja_to_hangle(sent):
    return hanja2hangle(sent, "substitution")


def only_hangle(sent):
    sent = hanja_to_hangle(sent)
    return doublespace_pattern.sub(" ", hangle_filter.sub(" ", sent)).strip()


def only_hangle_number(sent):
    sent = hanja_to_hangle(sent)
    return doublespace_pattern.sub(" ", hangle_number_filter.sub(" ", sent)).strip()


def only_text(sent):
    sent = hanja_to_hangle(sent)
    return doublespace_pattern.sub(" ", text_filter.sub(" ", sent)).strip()


def remain_hangle_on_last(eojeol):
    matchs = list(hangle_pattern.finditer(eojeol))
    if not matchs:
        return ""
    last_index = matchs[-1].span()[1]
    return eojeol[:last_index].strip()


def normalize_sent_for_lrgraph(sent):
    sent = hanja_to_hangle(sent)
    sent = text_filter_for_lrgraph.sub(" ", sent)
    sent = symbol_pattern.sub(" ", sent)
    sent_ = [remain_hangle_on_last(eojeol) for eojeol in sent.split()]
    sent_ = [eojeol for eojeol in sent_ if eojeol]
    if not sent_:
        return ""
    return " ".join(sent_)


class BaseNormalizer:
    """Abstract normalizer class from which all normalizers inherit.
    Subclasses must implement a ``normalize()`` method.
    """

    def normalize(self, text):
        """Normalize the text.
        :param string text: The text to normalize.
        :returns: Normalized text.
        :rtype: string
        """
        return text

    def __call__(self, text):
        """Calling a normalizer instance like a function just calls the normalize method."""
        return self.normalize(text)


class Normalizer(BaseNormalizer):
    """Main Normalizer class for generic English text.
    Normalize unicode, hyphens, quotes, whitespace.
    By default, the normal form NFKC is used for unicode normalization. This applies a compatibility decomposition,
    under which equivalent characters are unified, followed by a canonical composition. See Python docs for information
    on normal forms: http://docs.python.org/2/library/unicodedata.html#unicodedata.normalize
    """

    def __init__(
        self,
        ftfy={},
        spaces={},
        special_characters={},
        hanja2hangle=True,
        num_repeats=2,
        **kwargs
    ):
        """
        :param string form: Normal form for unicode normalization.
        :param bool strip: Whether to strip whitespace from start and end.
        :param bool hyphens: Whether to normalize all hyphens, minuses and dashes to the ASCII hyphen-minus character.
        :param bool quotes: Whether to normalize all apostrophes, quotes and primes to the ASCII quote character.
        :param bool ellipsis: Whether to normalize ellipses to three full stops.
        :param bool slashes: Whether to normalize slash characters to the ASCII slash character.
        :param bool tildes: Whether to normalize tilde characters to the ASCII tilde character.
        """
        self.ftfy = ftfy or {}
        self.special_characters = special_characters or {}
        self.spaces = spaces or {}

        self._uncurl_quotes = self.ftfy.get("uncurl_quotes", True)
        self._remove_control_chars = self.ftfy.get("remove_control_chars", True)

        self._strip = self.spaces.get("strip", True)
        self._fix_whitespaces = self.spaces.get("fix_whitespaces", True)
        self._collapse_whitespaces = self.spaces.get("collapse_whitespaces", True)
        self._replace_tabs = self.spaces.get("replace_tabs", True)
        self._replacement_spaces = " " * self.spaces.get("replacement_spaces", 4)

        self._fix_hyphens = self.special_characters.get("fix_hyphens", True)
        self._fix_ellipsis = self.special_characters.get("fix_ellipsis", True)
        self._fix_slashes = self.special_characters.get("fix_slashes", True)
        self._fix_tildes = self.special_characters.get("fix_tildes", True)
        self._fix_emoticons = self.special_characters.get("fix_emoticons", False)
        self._single_quotes_only = self.special_characters.get(
            "single_quotes_only", False
        )
        self._regular_parentheses_only = self.special_characters.get(
            "regular_parentheses_only", False
        )

        self._hanja2hangle = hanja2hangle
        self._num_repeats = num_repeats

        self._ftfy_cfg = TextFixerConfig(**self.ftfy)

    def fix_text(self, text):
        if text is None:
            return None
        return fix_text(str(text), self._ftfy_cfg)

    def remove_control_chars(self, text):
        """
        Strip out any control characters (they occasionally creep in somehow)
        """
        if text is None:
            return None
        for control in CONTROLS:
            text = text.replace(control, "")
        return text

    def fix_hyphens(self, text):
        """
        Normalize all hyphens, minuses and dashes to ascii hyphen-minus and remove soft hyphen entirely
        """
        # TODO: Better normalization of em/en dashes to '--' if surrounded by spaces or start/end?
        if text is None:
            return None
        for hyphen in HYPHENS | MINUSES:
            text = text.replace(hyphen, "-")
        text = text.replace("\u00ad", "")
        return text

    def uncurl_quotes(self, text):
        """
        Normalize all quotes and primes to ascii apostrophe and quotation mark
        """
        if text is None:
            return None
        for double_quote in DOUBLE_QUOTES:
            text = text.replace(double_quote, '"')  # \u0022
        for single_quote in SINGLE_QUOTES | APOSTROPHES | ACCENTS:
            text = text.replace(single_quote, "'")  # \u0027
        text = text.replace("′", "'")  # \u2032 prime
        text = text.replace("‵", "'")  # \u2035 reversed prime
        text = text.replace("″", "''")  # \u2033 double prime
        text = text.replace("‶", "''")  # \u2036 reversed double prime
        text = text.replace("‴", "'''")  # \u2034 triple prime
        text = text.replace("‷", "'''")  # \u2037 reversed triple prime
        text = text.replace("⁗", "''''")  # \u2057 quadruple prime
        return text

    def fix_ellipsis(self, text):
        """
        Normalize ellipses to three full stops
        """
        if text is None:
            return None
        text = text.replace("…", "...").replace(" . . . ", " ... ")  # \u2026
        return text

    def fix_slashes(self, text):
        """
        Normalize slash characters to ascii slash
        """
        if text is None:
            return None
        for slash in SLASHES:
            text = text.replace(slash, "/")
        return text

    def fix_tildes(self, text):
        """
        Normalize tilde characters to ascii tilde
        """
        if text is None:
            return None
        for tilde in TILDES:
            text = text.replace(tilde, "~")
        return text

    def replace_tabs(self, text, replacement_spaces=" " * 4):
        """
        Replace tabs with spaces
        """
        if text is None:
            return None
        text = text.replace("\t", replacement_spaces)
        return text

    def fix_whitespaces(self, text):
        """
        Normalize unusual whitespace not caught by unicodedata
        """
        if text is None:
            return None
        text = (
            text.replace("\u000b", " ").replace("\u000c", " ").replace(u"\u0085", " ")
        )
        return text

    def collapse_whitespaces(self, text):
        """
        Collapse all whitespace to a single space
        """
        if text is None:
            return None

        text = re.sub(r" +", " ", text)
        return text

    def single_quotes_only(self, text):
        """
        Replace all double quotes with single quotes
        """
        if text is None:
            return None
        for quote in QUOTES:
            text = text.replace(quote, "'")
        return text

    # Convert all brackets to regular parentheses
    def regular_parentheses_only(self, text):
        """
        Replace all curly brackets with regular parentheses
        """
        if text is None:
            return None
        for ob in LEFT_PARENTHESES:
            text = text.replace(ob, "(")
        for cb in RIGHT_PARENTHESES:
            text = text.replace(cb, ")")
        return text

    def hanja2hangle(self, text):
        """
        Convert all hanja to hangle
        """
        if text is None:
            return None
        text = hanja_to_hangle(text)
        return text

    def fix_emoticons(self, text, num_repeats=2):
        """
        Replace emoticons with their text equivalents
        """
        if text is None:
            return None
        text = emoticon_normalize(text, num_repeats=num_repeats)
        return text

    def normalize(self, text):
        """Run the Normalizer on a string.
        :param text: The string to normalize.
        """
        if text is None:
            return None
        text = self.fix_text(text)

        # Normalize to canonical unicode (using NFKC by default)
        # if self.form is not None:
        #     text = unicodedata.normalize(self.form, text)

        if self._remove_control_chars:
            text = self.remove_control_chars(text)

        # if self.fix_line_breaks:
        #     text = text.replace('\u2028', '\n').replace('\u2029', '\n').replace('\r\n', '\n').replace('\r', '\n')

        if self._fix_hyphens:
            text = self.fix_hyphens(text)

        if self._uncurl_quotes:
            text = self.uncurl_quotes(text)

        if self._fix_ellipsis:
            text = self.fix_ellipsis(text)

        if self._fix_slashes:
            text = self.fix_slashes(text)

        if self._fix_tildes:
            text = self.fix_tildes(text)

        if self._replace_tabs:
            text = self.replace_tabs(text, replacement_spaces=self._replacement_spaces)

        if self._fix_whitespaces:
            text = self.fix_whitespaces(text)

        if self._collapse_whitespaces:
            text = self.collapse_whitespaces(text)

        if self._strip:
            text = text.strip()

        if self._single_quotes_only:
            text = self.single_quotes_only(text)

        if self._regular_parentheses_only:
            text = self.regular_parentheses_only(text)

        if self._hanja2hangle:
            text = self.hanja2hangle(text)

        if self._fix_emoticons:
            text = self.emoticon_normalize(text, num_repeats=self._num_repeats)

        return text


#: Default normalize that canonicalizes unicode and fixes whitespace.
base_normalize = Normalizer(
    strip=True,
    fix_whitespaces=False,
    fix_hyphens=False,
    fix_ellipsis=False,
)
#: More aggressive normalize that also standardizes hyphens, and quotes.
strict_normalize = Normalizer(
    strip=True,
    fix_whitespaces=False,
    fix_hyphens=True,
    fix_ellipsis=True,
    fix_tildes=True,
)


# class ExcessNormalizer(Normalizer):
#     """Excessive string normalization.
#     This is useful when doing fuzzy string comparisons. A common use case is to run this before calculating the
#     Levenshtein distance between two strings, so that only "important" differences are counted.
#     """

#     def __init__(self, form='NFKC', strip=True, collapse=False, hyphens=True, quotes=True, ellipsis=True, tildes=True, lower=True):
#         """"""
#         super(ExcessNormalizer, self).__init__(form, strip=strip, fix_whitespaces=collapse, fix_hyphens=hyphens, uncurl_quotes=quotes,
#                                             fix_ellipsis=ellipsis, fix_tildes=tildes)
#         self.lower = lower

#     def normalize(self, text):
#         # Lowercase and normalize unicode
#         text = super(ExcessNormalizer, self).normalize(text.lower() if self.lower else text)
#         # Remove all whitespace
#         # text = ''.join(text.split())
#         # Convert all apostrophes, quotes, accents, primes to single ascii apostrophe
#         for quote in QUOTES:
#             text = text.replace(quote, "'")
#         # Convert all brackets to regular parentheses
#         for ob in {'(', '<', '[', '{', '&lt;'}:
#             text = text.replace(ob, '(')
#         for cb in {')', '>', ']', '}', '&gt;'}:
#             text = text.replace(cb, ')')
#         return text



"""Programmatic data set augmentation: TF creation and data generation utilities."""

from .apply.core import TFApplier  # noqa: F401
from .apply.pandas import PandasTFApplier  # noqa: F401
from .policy.core import ApplyAllPolicy, ApplyEachPolicy, ApplyOnePolicy  # noqa: F401
from .policy.sampling import MeanFieldPolicy, RandomPolicy  # noqa: F401
from .tf import TransformationFunction, transformation_function  # noqa: F401



# -*- encoding:utf8 -*-

import sys
import warnings
import re
import numpy as np

kor_begin = 44032
kor_end = 55203
chosung_base = 588
jungsung_base = 28
jaum_begin = 12593
jaum_end = 12622
moum_begin = 12623
moum_end = 12643

chosung_list = [
    "ㄱ",
    "ㄲ",
    "ㄴ",
    "ㄷ",
    "ㄸ",
    "ㄹ",
    "ㅁ",
    "ㅂ",
    "ㅃ",
    "ㅅ",
    "ㅆ",
    "ㅇ",
    "ㅈ",
    "ㅉ",
    "ㅊ",
    "ㅋ",
    "ㅌ",
    "ㅍ",
    "ㅎ",
]

jungsung_list = [
    "ㅏ",
    "ㅐ",
    "ㅑ",
    "ㅒ",
    "ㅓ",
    "ㅔ",
    "ㅕ",
    "ㅖ",
    "ㅗ",
    "ㅘ",
    "ㅙ",
    "ㅚ",
    "ㅛ",
    "ㅜ",
    "ㅝ",
    "ㅞ",
    "ㅟ",
    "ㅠ",
    "ㅡ",
    "ㅢ",
    "ㅣ",
]

jongsung_list = [
    " ",
    "ㄱ",
    "ㄲ",
    "ㄳ",
    "ㄴ",
    "ㄵ",
    "ㄶ",
    "ㄷ",
    "ㄹ",
    "ㄺ",
    "ㄻ",
    "ㄼ",
    "ㄽ",
    "ㄾ",
    "ㄿ",
    "ㅀ",
    "ㅁ",
    "ㅂ",
    "ㅄ",
    "ㅅ",
    "ㅆ",
    "ㅇ",
    "ㅈ",
    "ㅊ",
    "ㅋ",
    "ㅌ",
    "ㅍ",
    "ㅎ",
]

jaum_list = [
    "ㄱ",
    "ㄲ",
    "ㄳ",
    "ㄴ",
    "ㄵ",
    "ㄶ",
    "ㄷ",
    "ㄸ",
    "ㄹ",
    "ㄺ",
    "ㄻ",
    "ㄼ",
    "ㄽ",
    "ㄾ",
    "ㄿ",
    "ㅀ",
    "ㅁ",
    "ㅂ",
    "ㅃ",
    "ㅄ",
    "ㅅ",
    "ㅆ",
    "ㅇ",
    "ㅈ",
    "ㅉ",
    "ㅊ",
    "ㅋ",
    "ㅌ",
    "ㅍ",
    "ㅎ",
]

moum_list = [
    "ㅏ",
    "ㅐ",
    "ㅑ",
    "ㅒ",
    "ㅓ",
    "ㅔ",
    "ㅕ",
    "ㅖ",
    "ㅗ",
    "ㅘ",
    "ㅙ",
    "ㅚ",
    "ㅛ",
    "ㅜ",
    "ㅝ",
    "ㅞ",
    "ㅟ",
    "ㅠ",
    "ㅡ",
    "ㅢ",
    "ㅣ",
]

doublespace_pattern = re.compile(r"\s+")
repeatchars_pattern = re.compile(r"(\w)\\1{3,}")


def normalize(
    doc, english=False, number=False, punctuation=False, remove_repeat=0, remains={}
):
    message = "normalize func will be moved soynlp.normalizer at ver 0.1\nargument remains will be removed at ver 0.1"
    warnings.warn(message, DeprecationWarning, stacklevel=2)

    if remove_repeat > 0:
        doc = repeatchars_pattern.sub(r"\\1" * remove_repeat, doc)

    f = []
    for c in doc:
        if c == " ":
            f.append(c)
            continue
        i = to_base(c)
        if (
            (kor_begin <= i <= kor_end)
            or (jaum_begin <= i <= jaum_end)
            or (moum_begin <= i <= moum_end)
        ):
            f.append(c)
            continue
        if (english) and ((i >= 97 and i <= 122) or (i >= 65 and i <= 90)):
            f.append(c)
            continue
        if (number) and (i >= 48 and i <= 57):
            f.append(c)
            continue
        if (punctuation) and (
            i == 33 or i == 34 or i == 39 or i == 44 or i == 46 or i == 63 or i == 96
        ):
            f.append(c)
            continue
        if c in remains:
            f.append(c)
            continue
        else:
            f.append(" ")
    return doublespace_pattern.sub(" ", "".join(f)).strip()


def compose(chosung, jungsung, jongsung):
    return chr(
        kor_begin
        + chosung_base * chosung_list.index(chosung)
        + jungsung_base * jungsung_list.index(jungsung)
        + jongsung_list.index(jongsung)
    )


def decompose(c):
    if not character_is_korean(c):
        return None
    i = to_base(c)
    if jaum_begin <= i <= jaum_end:
        return (c, " ", " ")
    if moum_begin <= i <= moum_end:
        return (" ", c, " ")
    i -= kor_begin
    cho = i // chosung_base
    jung = (i - cho * chosung_base) // jungsung_base
    jong = i - cho * chosung_base - jung * jungsung_base
    return (chosung_list[cho], jungsung_list[jung], jongsung_list[jong])


def character_is_korean(c):
    i = to_base(c)
    return (
        (kor_begin <= i <= kor_end)
        or (jaum_begin <= i <= jaum_end)
        or (moum_begin <= i <= moum_end)
    )


def character_is_complete_korean(c):
    return kor_begin <= to_base(c) <= kor_end


def character_is_jaum(c):
    return jaum_begin <= to_base(c) <= jaum_end


def character_is_moum(c):
    return moum_begin <= to_base(c) <= moum_end


def to_base(c):
    if type(c) == str or type(c) == int:
        return ord(c)
    else:
        raise TypeError


def character_is_number(i):
    i = to_base(i)
    return i >= 48 and i <= 57


def character_is_english(i):
    i = to_base(i)
    return (i >= 97 and i <= 122) or (i >= 65 and i <= 90)


def character_is_punctuation(i):
    i = to_base(i)
    return i == 33 or i == 34 or i == 39 or i == 44 or i == 46 or i == 63 or i == 96


class ConvolutionHangleEncoder:
    """초/중/종성을 구성하는 자음/모음과 띄어쓰기만 인코딩
    one hot vector [ㄱ, ㄴ, ㄷ, ... ㅎ, ㅏ, ㅐ, .. ㅢ, ㅣ,"  ", ㄱ, ㄲ, ... ㅍ, ㅎ,"  ", 0, 1, 2, .. 9]
    """

    def __init__(self):
        self.jung_begin = 19  # len(chosung_list)
        self.jong_begin = 40  # self.jung_begin + len(jungsung_list)
        self.number_begin = 68  # self.jong_begin + len(jongsung_list)
        self.space = (
            78  # len(chosung_list) + len(jungsung_list) + len(jongsung_list) + 10
        )
        self.unk = 79
        self.dim = 80
        num = [str(i) for i in range(10)]
        space = " "
        unk = "<unk>"
        idx_to_char = (
            chosung_list + jungsung_list + jongsung_list + num + [space] + [unk]
        )
        self.idx_to_char = np.asarray(idx_to_char)
        self.jamo_to_idx = {
            "ㄱ": 0,
            "ㄲ": 1,
            "ㄴ": 2,
            "ㄷ": 3,
            "ㄸ": 4,
            "ㄹ": 5,
            "ㅁ": 6,
            "ㅂ": 7,
            "ㅃ": 8,
            "ㅅ": 9,
            "ㅆ": 10,
            "ㅇ": 11,
            "ㅈ": 12,
            "ㅉ": 13,
            "ㅊ": 14,
            "ㅋ": 15,
            "ㅌ": 16,
            "ㅍ": 17,
            "ㅎ": 18,
            "ㅏ": 19,
            "ㅐ": 20,
            "ㅑ": 21,
            "ㅒ": 22,
            "ㅓ": 23,
            "ㅔ": 24,
            "ㅕ": 25,
            "ㅖ": 26,
            "ㅗ": 27,
            "ㅘ": 28,
            "ㅙ": 29,
            "ㅚ": 30,
            "ㅛ": 31,
            "ㅜ": 32,
            "ㅝ": 33,
            "ㅞ": 34,
            "ㅟ": 35,
            "ㅠ": 36,
            "ㅡ": 37,
            "ㅢ": 38,
            "ㅣ": 39,
            " ": 40,
            "ㄳ": 43,
            "ㄵ": 45,
            "ㄶ": 46,
            "ㄺ": 49,
            "ㄻ": 50,
            "ㄼ": 51,
            "ㄽ": 52,
            "ㄾ": 53,
            "ㄿ": 54,
            "ㅀ": 55,
            "ㅄ": 58,
        }

    def encode(self, sent):
        onehot = self.sent_to_onehot(sent)
        x = np.zeros((len(onehot), self.dim))
        for i, xi in enumerate(onehot):
            for j in xi:
                x[i, j] = 1
        return x

    def sent_to_onehot(self, sent):
        chars = self._normalize(sent)
        ords = [ord(c) for c in chars]
        onehot = []
        for char, idx in zip(chars, ords):
            if idx == 32:
                onehot.append((self.space,))
            elif 48 <= idx <= 57:
                onehot.append((idx - 48 + self.number_begin,))
            else:
                onehot.append(self._decompose(char, idx))
        return onehot

    def onehot_to_sent(self, encoded_sent):
        def check_cjj(c):
            cho, jung, jong = c
            if not (0 <= cho < self.jung_begin):
                raise ValueError("Chosung %d is out of index" % cho)
            if not (self.jung_begin <= jung < self.jong_begin):
                raise ValueError("Jungsung %d is out of index" % jung)
            if not (self.jong_begin <= jong < self.number_begin):
                raise ValueError("Jongsung %d is out of index" % jong)

        chars = []
        for c in encoded_sent:
            if len(c) == 1:
                if not 0 <= c[0] < self.dim:
                    raise ValueError(
                        "character index %d is out of index [0, %d]" % (c[0], self.dim)
                    )
                chars.append(self.idx_to_char[c[0]])
            elif len(c) == 3:
                check_cjj(c)
                cho, jung, jong = tuple(self.idx_to_char[ci] for ci in c)
                chars.append(compose(cho, jung, jong))
            else:
                chars.append(self.idx_to_char[-1])
        return "".join(chars)

    def _normalize(self, sent):
        import re

        regex = re.compile("[^ㄱ-ㅎㅏ-ㅣ가-힣 0-9]")
        sent = regex.sub(" ", sent)
        sent = doublespace_pattern.sub(" ", sent).strip()
        return sent

    def _compose(self, cho, jung, jong):
        return chr(kor_begin + chosung_base * cho + jungsung_base * jung + jong)

    def _decompose(self, c, i):
        if kor_begin <= i <= kor_end:
            i -= kor_begin
            cho = i // chosung_base
            jung = (i - cho * chosung_base) // jungsung_base
            jong = i - cho * chosung_base - jung * jungsung_base
            return (cho, self.jung_begin + jung, self.jong_begin + jong)
        else:
            return (self.jamo_to_idx.get(c, self.unk),)



import logging
from collections import defaultdict
from typing import DefaultDict, Dict, List, Optional


class Logger:
    """Class for logging LabelModel.

    Parameters
    ----------
    log_freq
        Number of units at which to log model

    Attributes
    ----------
    log_freq
        Number of units at which to log model
    unit_count
        Running total of number of units passed without logging
    """

    def __init__(self, log_freq: int) -> None:
        self.log_freq = log_freq
        self.unit_count = -1

    def check(self) -> bool:
        """Check if the logging frequency has been met.

        Returns
        -------
        bool
            Whether to log or not based on logging frequency
        """
        self.unit_count += 1
        return self.unit_count % self.log_freq == 0

    def log(self, metrics_dict: Dict[str, float]) -> None:
        """Print all metrics in metrics_dict to screen.

        Parameters
        ----------
        metrics_dict
            Dictionary of metric names (keys) and values to log

        Raises
        ------
        Exception
            If metric names formatted incorrectly
        """
        score_strings: DefaultDict[str, List[str]] = defaultdict(list)
        for full_name, value in metrics_dict.items():
            task: Optional[str]
            if full_name.count("/") == 2:
                task, split, metric = full_name.split("/")
            elif full_name.count("/") == 1:
                task = None
                split, metric = full_name.split("/")
            else:
                msg = f"Metric should have form task/split/metric or split/metric, not: {full_name}"
                raise Exception(msg)

            if task:
                metric_name = f"{task}/{metric}"
            else:
                metric_name = metric
            if isinstance(value, float):
                score_strings[split].append(f"{metric_name}={value:0.3f}")
            else:
                score_strings[split].append(f"{metric_name}={value}")

        string = f"[{self.unit_count} epochs]:"

        if score_strings["train"]:
            train_scores = f"{', '.join(score_strings['train'])}"
            string += f" TRAIN:[{train_scores}]"
        if score_strings["valid"]:
            valid_scores = f"{', '.join(score_strings['valid'])}"
            string += f" VALID:[{valid_scores}]"
        logging.info(string)



import os
import glob
import orjson as json
from pathlib import Path
from hydra.utils import instantiate
from ekorpkit import eKonf

from ..utils.convert import convert_tf_checkpoint_to_pytorch


def convert_electra_ckpt_to_pytorch(**cfg):
    cfg = eKonf.to_config(cfg)

    model_config = eKonf.to_dict(cfg.model.electra, resolve=True)
    args = cfg.task.train.convert_electra
    tokenizer_config = dict(args.tokenizer)

    print(f"Converting tensorflow checkpoints to pytoch")
    os.makedirs(args.torch_output_dir, exist_ok=True)
    ckpt_file = f"{args.checkpoints_dir}/checkpoint"
    org_contents = open(ckpt_file, "r").read()

    if args.checkpoint_files:
        if isinstance(args.checkpoint_files, str):
            args.checkpoint_files = [args.checkpoint_files]
        checkpoint_files = args.checkpoint_files
    else:
        checkpoint_files = [
            Path(f).stem for f in glob.glob(f"{args.checkpoints_dir}/*.index")
        ]

    for ckpt in checkpoint_files:

        print("processing checkpoint: {}".format(ckpt))
        ckpt_contents = f'model_checkpoint_path: "{ckpt}"\n'
        open(ckpt_file, "w").write(ckpt_contents)
        for model_type in ["discriminator", "generator"]:
            torch_config = model_config[model_type]
            output_dir = f"{args.torch_output_dir}/{ckpt}/{model_type}"
            os.makedirs(output_dir, exist_ok=True)
            # create config
            config_file = f"{output_dir}/config.json"
            with open(config_file, "w") as f:
                json.dump(torch_config, f, ensure_ascii=False, indent=4)
            with open(f"{output_dir}/tokenizer_config.json", "w") as f:
                json.dump(tokenizer_config, f, ensure_ascii=False, indent=4)
            # copy vocab
            os.system(f"cp -f {args.vocab_file} {output_dir}/vocab.txt")
            # model path
            pytorch_dump_path = f"{output_dir}/pytorch_model.bin"
            print(f"Saving {model_type} to {pytorch_dump_path}")
            convert_tf_checkpoint_to_pytorch(
                args.checkpoints_dir,
                config_file,
                pytorch_dump_path,
                model_type,
                args.load_weights_func._target_,
            )
    # restore original contents
    open(ckpt_file, "w").write(org_contents)



import sys
import re
import requests
from bs4 import BeautifulSoup
from datetime import datetime
from .base import FOMC


class Statement(FOMC):
    """
    A convenient class for extracting statement from the FOMC website
    """

    def __init__(self, content_type, **args):
        super().__init__(content_type, **args)

    def _get_links(self, from_year):
        """
        Override private function that sets all the links for the contents to download on FOMC website
         from from_year (=min(2015, from_year)) to the current most recent year
        """
        self.links = []
        self.titles = []
        self.speakers = []
        self.dates = []

        r = requests.get(self.calendar_url)
        soup = BeautifulSoup(r.text, "html.parser")

        # Getting links from current page. Meetin scripts are not available.
        if self.verbose:
            print("Getting links for statements...")
        contents = soup.find_all(
            "a", href=re.compile("^/newsevents/pressreleases/monetary\d{8}[ax].htm")
        )
        self.links = [content.attrs["href"] for content in contents]
        self.speakers = [
            self._speaker_from_date(self._date_from_link(x)) for x in self.links
        ]
        self.titles = ["FOMC Statement"] * len(self.links)
        self.dates = [
            datetime.strptime(self._date_from_link(x), "%Y-%m-%d") for x in self.links
        ]
        # Correct some date in the link does not match with the meeting date
        for i, m_date in enumerate(self.dates):
            if m_date == datetime(2019, 10, 11):
                self.dates[i] = datetime(2019, 10, 4)

        if self.verbose:
            print("{} links found in the current page.".format(len(self.links)))

        # Archived before 2015
        if from_year <= 2014:
            print("Getting links from archive pages...")
            for year in range(from_year, 2015):
                yearly_contents = []
                fomc_yearly_url = (
                    self.base_url
                    + "/monetarypolicy/fomchistorical"
                    + str(year)
                    + ".htm"
                )
                r_year = requests.get(fomc_yearly_url)
                soup_yearly = BeautifulSoup(r_year.text, "html.parser")
                yearly_contents = soup_yearly.findAll("a", text="Statement")
                for yearly_content in yearly_contents:
                    self.links.append(yearly_content.attrs["href"])
                    self.speakers.append(
                        self._speaker_from_date(
                            self._date_from_link(yearly_content.attrs["href"])
                        )
                    )
                    self.titles.append("FOMC Statement")
                    self.dates.append(
                        datetime.strptime(
                            self._date_from_link(yearly_content.attrs["href"]),
                            "%Y-%m-%d",
                        )
                    )
                    # Correct some date in the link does not match with the meeting date
                    if self.dates[-1] == datetime(2007, 6, 18):
                        self.dates[-1] = datetime(2007, 6, 28)
                    elif self.dates[-1] == datetime(2007, 8, 17):
                        self.dates[-1] = datetime(2007, 8, 16)
                    elif self.dates[-1] == datetime(2008, 1, 22):
                        self.dates[-1] = datetime(2008, 1, 21)
                    elif self.dates[-1] == datetime(2008, 3, 11):
                        self.dates[-1] = datetime(2008, 3, 10)
                    elif self.dates[-1] == datetime(2008, 10, 8):
                        self.dates[-1] = datetime(2008, 10, 7)

                if self.verbose:
                    print(
                        "YEAR: {} - {} links found.".format(year, len(yearly_contents))
                    )

        print("There are total ", len(self.links), " links for ", self.content_type)

    def _add_article(self, link, index=None):
        """
        Override a private function that adds a related article for 1 link into the instance variable
        The index is the index in the article to add to.
        Due to concurrent processing, we need to make sure the articles are stored in the right order
        """
        if self.verbose:
            sys.stdout.write(".")
            sys.stdout.flush()

        res = requests.get(self.base_url + link)
        html = res.text
        article = BeautifulSoup(html, "html.parser")
        paragraphs = article.findAll("p")
        para_text = []
        for paragraph in paragraphs:
            text = re.sub(r"\s+", " ", paragraph.get_text().strip())
            if len(text) > 0 and text not in self.skip_text_list:
                para_text.append(text)
        self.articles[index] = self.segment_separator.join(para_text)



from typing import Any, Dict

from ..types import Config


def merge_config(config: Config, config_updates: Dict[str, Any]) -> Config:
    """Merge a (potentially nested) dict of kwargs into a config (NamedTuple).

    Parameters
    ----------
    config
        An instantiated Config to update
    config_updates
        A potentially nested dict of settings to update in the Config

    Returns
    -------
    Config
        The updated Config

    Example
    -------
    ```
    config_updates = {
        "n_epochs": 5,
        "optimizer_config": {
            "lr": 0.001,
        }
    }
    trainer_config = merge_config(TrainerConfig(), config_updates)
    ```
    """
    for key, value in config_updates.items():
        if isinstance(value, dict):
            config_updates[key] = merge_config(getattr(config, key), value)
    return config._replace(**config_updates)
